@@ -5,8 +5,8 @@
 # Author            : Sébastien Pierre
 # License           : BSD License
 # -----------------------------------------------------------------------------
-# Creation date     : 2014-DEC-18
-# Last modification : 2014-DEC-23
+# Creation date     : 18-Dec-2014
+# Last modification : 24-Dec-2014
 # -----------------------------------------------------------------------------
 
 import sys, os, re
@@ -18,7 +18,7 @@ try:
 except ImportError:
 	import logging
 
-VERSION            = "0.3.3"
+VERSION            = "0.4.0"
 LICENSE            = "http://ffctn.com/doc/licenses/bsd"
 PACKAGE_PATH       = dirname(abspath(__file__))
 LIB_PATH           = dirname(PACKAGE_PATH)
@@ -47,9 +47,11 @@ if os.path.exists(H_PATH):
 		"typedef char* iterated_t;\n"
 		"typedef void Element;\n"
 		"typedef struct ParsingElement ParsingElement;\n"
-		"typedef struct ParsingElement ParsingElement;\n"
+		"typedef struct ParsingResult  ParsingResult;\n"
+		"typedef struct ParsingStats   ParsingStats;\n"
 		"typedef struct ParsingContext ParsingContext;\n"
 		"typedef struct Match Match;\n"
+		"typedef struct Grammar Grammar;\n"
 		"typedef struct TokenMatchGroup TokenMatchGroup;\n"
 	) + clib.getCode(
 		("ConditionCallback",    None),
@@ -61,6 +63,8 @@ if os.path.exists(H_PATH):
 		("Iterator*",            O),
 		("ParsingContext*",      O),
 		("ParsingElement*",      O),
+		("ParsingResult*",       O),
+		("ParsingStats*",        O),
 		("Word*" ,               O),
 		("Token",                O),
 		("TokenMatch",           O),
@@ -81,7 +85,7 @@ else:
 ffi = FFI()
 ffi.cdef(cdef)
 lib = None
-for p in (LIB_PATH, LIB_ALT_PATH):
+for p in (LIB_PATH, LIB_ALT_PATH, "."):
 	p = join(p, "libparsing.so")
 	if os.path.exists(p):
 		lib = ffi.dlopen(p)
@@ -517,6 +521,33 @@ class Symbols:
 	def __getitem__( self, key ):
 		return getattr(self, key)
 
+# -----------------------------------------------------------------------------
+#
+# PARSING RESULT
+#
+# -----------------------------------------------------------------------------
+
+class ParsingResult(CObject):
+
+	_TYPE = "ParsingResult*"
+
+	def match( self ):
+		return Match.Wrap(self._cobject.match)
+
+	def status( self ):
+		return self._cobject.status
+
+	def offset( self ):
+		return self._cobject.context.iterator.offset
+
+	def text( self ):
+		return ffi.string(self._cobject.iterator.buffer)
+
+	def __del__( self ):
+		# The parsing result is the only one we really need to free
+		# along with the grammar
+		lib.ParsingResult_free(self._cobject)
+
 # -----------------------------------------------------------------------------
 #
 # GRAMMAR
@@ -525,10 +556,12 @@ class Symbols:
 
 class Grammar(CObject):
 
-	def _new(self, name=None ):
+	def _new(self, name=None, verbose=True ):
 		self.name    = name
 		self.symbols = Symbols()
-		return lib.Grammar_new()
+		g =  lib.Grammar_new()
+		g.isVerbose = 1 if verbose else 0
+		return g
 
 	def list( self ):
 		res = []
@@ -603,7 +636,10 @@ class Grammar(CObject):
 		return self
 
 	def parsePath( self, path ):
-		return Match(lib.Grammar_parseFromPath(self._cobject, path))
+		return ParsingResult.Wrap(lib.Grammar_parseFromPath(self._cobject, path))
+
+	def parseString( self, text ):
+		return ParsingResult.Wrap(lib.Grammar_parseFromString(self._cobject, text))
 
 	def walk( self, callback ):
 		def c(o,s):
@@ -617,6 +653,11 @@ class Grammar(CObject):
 		c = ffi.callback("int(*)(void *, int)", c)
 		return lib.Element__walk(self._cobject.axiom, c, 0)
 
+	def __del__( self ):
+		# The parsing result is the only one we really need to free
+		# along with the grammar
+		lib.Grammar_free(self._cobject)
+
 # -----------------------------------------------------------------------------
 #
 # ABSTRACT PROCESSOR
@@ -657,6 +698,7 @@ class AbstractProcessor:
 	def process( self, match ):
 		# res = [self.walk(_) for _ in match.children()]
 		#print "MATCH", match.element().name(), ":", res, self._process(match)
+		if isinstance(match, ParsingResult): match = match.match()
 		return self._process(match)
 
 	def _process( self, match ):
